#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''Run as Test-case to generate results for sample-vehicles.
Run it as cmd-line to compare with Heinz's results.

:created: 5 Jan 2014
'''

from matplotlib import pyplot as plt
from wltp.experiment import Experiment
from wltp.test.goodvehicle import goodVehicle
from wltp.experiment import applyDriveabilityRules
import glob
import logging
import math
import numpy as np
import numpy.testing as npt
import os
import pandas as pd
import re
import unittest


mydir = os.path.dirname(__file__)
samples_folder = 'samples'
loglevel = level=logging.DEBUG
logging.basicConfig(level=loglevel)
logging.getLogger().setLevel(loglevel)
log = logging.getLogger(__name__)

class ExperimentSampleVehs(unittest.TestCase):
    '''Compares a batch of vehicles with results obtained from "Official" implementation.'''


    def setUp(self):
        self.run_comparison = True # NOTE: Set 'False' to UPDATE sample-results or run main() (assuming they are ok).


    def testSampleVehicles(self, plot_results=False, encoding="ISO-8859-1"):
        run_the_experiments(plot_results=False, compare_results=self.run_comparison, encoding="ISO-8859-1")


def run_the_experiments(transplant_original_gears=False, plot_results=False, compare_results=False, encoding="ISO-8859-1"):
    # rated_power,kerb_mass,rated_speed,idling_speed,test_mass,no_of_gears,ndv_1,ndv_2,ndv_3,ndv_4,ndv_5,ndv_6,ndv_7,ID_cat,user_def_driv_res_coeff,user_def_power_curve,f0,f1,f2,Comment
    # 0                                                            5                                10                                                    15                        19
    csvfname = 'sample_vehicles.csv'
    csvfname = os.path.join(mydir, samples_folder, csvfname)
    df = pd.read_csv(csvfname, encoding = encoding, index_col = 0)

    for (ix, row) in df.iterrows():
        veh_num = ix

        model = goodVehicle()
        veh = model['vehicle']

        veh['mass'] = row['test_mass']
        veh['resistance_coeffs'] = list(row['f0':'f2'])
        veh['p_rated'] = row['rated_power']
        veh['n_rated'] = row['rated_speed']
        veh['n_idle'] = int(row['idling_speed'])
        ngears = int(row['no_of_gears'])
        veh['gear_ratios'] = list(row[6:6+ngears]) #'ndv_1'

        experiment = Experiment(model)
        model = experiment.run()

        if (transplant_original_gears):
            log.warning(">>> Transplanting gears from Heinz's!")
            cycle_run = model['cycle_run']
            hz_df = read_heinz_file(veh_num)
            GEARS = np.array(hz_df['g_max'])
            cycle_run['gears_orig'] = GEARS.copy()

            V = np.array(cycle_run['v_class'])
            A = np.append(np.diff(V), 0)
            CLUTCH = np.array(cycle_run['clutch'])
            driveability_issues = np.empty_like(V, dtype=object)
            driveability_issues[:] = ''
            applyDriveabilityRules(V, A, GEARS, CLUTCH, len(veh['gear_ratios']), driveability_issues)

            cycle_run['gears'] = GEARS

        params = model['params']

        f_downscale = params['f_downscale']
        if (f_downscale > 0):
            log.warning('>> DOWNSCALE %s', f_downscale)


        # ankostis_mdb:  't', "v in km/h","v_orig","a in m/s²","gear","g_min","g_max","gear_modification","error_description"
        # heinz:         't', 'km_h', 'stg', 'gear'

        (root, ext) = os.path.splitext(csvfname)
        outfname = '{}-{:05}{}'.format(root, veh_num, ext)
        df = pd.DataFrame(model['cycle_run'])

        compare_exp_results(df, outfname, compare_results)
        df.to_csv(outfname, index_label='time')


def read_sample_file(inpfname):
    df = pd.read_csv(inpfname)

    return df


def read_heinz_file(veh_num, heinz_dir=None):
    if (heinz_dir is None):
        heinz_dir = os.path.join(mydir, samples_folder)

    vehfpath = os.path.join(heinz_dir, 'heinz_Petrol_veh{:05}.csv'.format(veh_num))
    try:
        inpfname = glob.glob(vehfpath)[0]
    except IndexError:
        raise FileNotFoundError(vehfpath)
    df = pd.read_csv(inpfname, encoding='latin-1', header=0, index_col=3)
    assert df['vehicle_no'][0] == veh_num

#     vehfpath = os.path.join(samples_folder, 'heinz_Petrol_veh{:05}.dri'.format(veh_num))
#     inpfname = glob.glob(vehfpath)[0]
#     df = pd.read_csv(inpfname, encoding='latin-1')

    return df


###################
# COMPARE RESULTS #
###################

def compare_exp_results(tabular, outfname, run_comparison):
    if (run_comparison):
        try:
            data_prev = read_sample_file(outfname)
            ## Compare changed-tabular
            #
            npt.assert_array_equal(tabular['gears'],  data_prev['gears'])
            # Unreached code in case of assertion.
            # cmp = tabular['gears'] != data_prev['gears']
            # if (cmp.any()):
            #     self.plotResults(data_prev)
            #     print('>> COMPARING(%s): %s'%(fname, cmp.nonzero()))
            # else:
            #     print('>> COMPARING(%s): OK'%fname)
        except OSError as ex:  # @UnusedVariable
            print('>> COMPARING(%s): No old-tabular found, 1st time to run' % outfname)
            run_comparison = False


def plotResults(veh_fname, my_df, hz_df,  g_diff, ax, plot_diffs_gears_only=True, plot_original_gears = False):
    if (plot_original_gears):
        my_gear_col = 'gears_orig'
        hz_gear_col = 'g_max'
    else:
        my_gear_col = 'gears'
        hz_gear_col = 'gear'

    ax.grid(True)

    ax2 = ax.twinx()

    tlen = len(my_df.index)
    #ax.set_xticks(np.arange(0.0, tlen, 50.0)) NO! looses auto when zooming.


    clutch = my_df['clutch']
    clutch = clutch.nonzero()[0]
    ax.vlines(clutch,  0, 0.2)

    ## Add pickers on driveability lines showing the specific msg.
    #
    driveability = my_df['driveability']
    driveability_true = driveability.apply(lambda s: isinstance(s, str))
    lines = ax2.vlines(driveability_true.nonzero()[0],  2, 4, 'c', picker=5)
    lines.set_urls(driveability[driveability_true])

    v_max = my_df['v_class'].max()
    ax.hlines(1 / v_max,  0, tlen, color="0.75")

    ax.plot(my_df['v_class'] / v_max)
    ax.plot(my_df['v_target'] / v_max, '-.')

#     ax.plot(my_df['rpm'] / my_df['rpm'].max())
#     p_req = my_df['p_required'] / my_df['p_required'].max()
#     p_req[p_req < 0] = 0
#     ax.plot(p_req)

    ## Plot gear diffs.
    #
    my_gears = my_df[my_gear_col]
    hz_v_real = hz_df['v']
    hz_v_target = hz_df['v_downscale']
    hz_gears = hz_df[hz_gear_col]


    orig_gears = my_df['gears_orig']
    if plot_diffs_gears_only:
        diff_gears = my_gears != hz_gears
        difft = diff_gears.nonzero()[0]
        difft = set().union(difft,
                    difft+1, difft+2, difft+3, difft+4, difft+5, difft+6,
                    difft-1, difft-2, difft-3, difft-4, difft-5, difft-6,
        )
        difft = list(difft)
        my_gears = my_gears[difft]
        hz_gears = hz_gears[difft]
        ax2.plot(difft, my_gears.tolist(), 'o', color='red')
        ax2.plot(difft, orig_gears[difft].tolist(), 'v', color='green')
        ax2.plot(difft, hz_gears.tolist(), '*', color='blue')
    else:
        ax2.plot(my_gears.tolist(), 'o', color='red')
        ax2.plot(hz_gears.tolist(), '*', color='blue')



    ax.plot(my_df['v_real'] / v_max)

    ## Add pickers on driveability lines showing the specific msg.
    #
    hz_driveability = hz_df['gear_modification']
    hz_driveability_true = ~hz_driveability.apply(np.isreal)
    lines = ax2.vlines(hz_driveability_true.nonzero()[0],  0, 2, 'm', picker=5)
    lines.set_urls(hz_driveability[hz_driveability_true])


    ax.plot(hz_v_target / v_max, '--')
    ax.plot(hz_v_real / v_max, ':')

    ax.text(0.7, 0, 'Diffs: %.4f' % g_diff, transform=ax.transAxes, bbox={'facecolor':'red', 'alpha':0.5, 'pad':10})


def plot_diffs_with_heinz(heinz_dir, experiment_num=None, transplant_original_gears=False, force_rerun_experiments=False):
    def is_experiments_outdated(outfiles):
        if not outfiles or force_rerun_experiments:
            return True

        resfiles_date = min([os.path.getmtime(file) for file in outfiles])
        checkfiles = ['../model.py', '../experiment.py', 'samples/sample_vehicles.csv']
        checkdates = [os.path.getmtime(os.path.join(mydir, fheinz_fname)) for fheinz_fname in checkfiles]
        modifs = [fdate > resfiles_date for fdate in checkdates]
        return any(modifs)



    def read_and_compare_experiment(samples_folder, myfname, veh_num):
        df_my = read_sample_file(myfname)
        df_hz = read_heinz_file(veh_num, samples_folder)

        ## Count base-calc errors (before dirveability).
        ndiff_gears_orig = np.count_nonzero(df_my['gears_orig'] != df_hz['g_max'])

        ## Count all errors.
        #
        my_gears = df_my['gears']
        gears_hz = df_hz['gear']
        diff_gears = (my_gears != gears_hz)
        ndiff_gears = np.count_nonzero(diff_gears)

        ## Count Acceleration-only errors.
        #
        accel = np.gradient(df_my['v_class'])
        diff_gears_accel = diff_gears[accel >= 0]
        ndiff_gears_accel = np.count_nonzero(diff_gears_accel)

        return (df_my, df_hz, ndiff_gears, ndiff_gears_accel, ndiff_gears_orig)



    fig = plt.figure()
    text_infos = fig.text(0.5, 0.5, '', transform=fig.transFigure, bbox={'facecolor':'grey', 'alpha':0.4, 'pad':10}, horizontalalignment='center', verticalalignment='center', color='blue')

    def fig_onpick(event):
        pickline = event.artist
        urls = pickline.get_urls()
        rule = urls.iloc[event.ind]
        print(rule)
        text_infos.set_text('Rule: %s' % rule)

        fig.canvas.draw()

    fig.canvas.mpl_connect('pick_event', fig_onpick)


    if experiment_num is None:
        paths = glob.glob(os.path.join(mydir, samples_folder, 'sample_vehicles-*.csv'))

        if is_experiments_outdated(paths):
            run_the_experiments(transplant_original_gears = transplant_original_gears)
            paths = glob.glob(os.path.join(mydir, samples_folder, 'sample_vehicles-*.csv'))

        npaths          = len(paths)

        ## NOTE: Limit subplots to facilitate research.
        #
#         paths_to_plot = paths
#         paths_to_plot = paths[0:9]
#         paths_to_plot = paths[5:6] + paths[7:9] + paths[14:16] + paths[23:24]
        paths_to_plot = paths[5:8] + paths[17:18] + paths[22:24]

        ## Decide subplot-grid dimensions.
        #
        npaths_to_plot  = len(paths_to_plot)
        w = math.ceil(math.sqrt(npaths_to_plot))
        h = w-1 if ((w-1) * w >= npaths_to_plot) else w

        g_diff = np.zeros((3, npaths))
        nplotted = 0
        for (n, inpfname) in enumerate(paths):
            m = re.match('.*sample_vehicles-(\d+).csv', inpfname)
            assert m


            veh_num = int(m.group(1))
            (df_my, df_hz, ndiff_gears, ndiff_gears_accel, ndiff_gears_orig) \
                                    = read_and_compare_experiment(heinz_dir, inpfname, veh_num)
            g_diff[0, n]            = ndiff_gears
            g_diff[1, n]            = ndiff_gears_accel
            g_diff[2, n]            = ndiff_gears_orig

            log.info(">> %i: %s: ±DIFFs(%i), +DIFFs(%i), ±ORIGs(%i)", n, inpfname, ndiff_gears, ndiff_gears_accel, ndiff_gears_orig)


            if (inpfname in paths_to_plot):
                nplotted += 1
                ax = fig.add_subplot(w, h, nplotted)
                veh_name = os.path.basename(inpfname)
                ax.set_title('%i: %s'%(n, veh_name), fontdict={'fontsize': 8} )
                plotResults(veh_name, df_my, df_hz, ndiff_gears, ax, plot_original_gears = not transplant_original_gears)

        orig = 'Driveability' if transplant_original_gears else 'Pre-Driveability'
        fig.suptitle('%s: ±DIFFs: count(%i), min(%i), MEAN(%.2f±%.2f), max(%i).' % (orig, g_diff[0].sum(), g_diff[0].min(), g_diff[0].mean(), g_diff[0].std(), g_diff[0].max()))
        log.info('#       ±DIFFs: count(%i), min(%i), MEAN(%.2f±%.2f), max(%i).', g_diff[0].sum(), g_diff[0].min(), g_diff[0].mean(), g_diff[0].std(), g_diff[0].max())
        log.info('#       +DIFFs: count(%i), min(%i), MEAN(%.2f±%.2f), max(%i).', g_diff[1].sum(), g_diff[1].min(), g_diff[1].mean(), g_diff[1].std(), g_diff[1].max())
        log.info('#       ±ORIGs: count(%i), min(%i), MEAN(%.2f±%.2f), max(%i).', g_diff[2].sum(), g_diff[2].min(), g_diff[2].mean(), g_diff[2].std(), g_diff[2].max())

        ## RESULTS:
        #    0.0.5-alpha:                 min(0.6108%), MEAN(3.6276%), max(12.1599%)
        #    0.0.5-alpha:    count(1960), min(11.0000), MEAN(65.3333), max(219.0000)
        #    ±Accel erros    count(6486), min(151.0000), MEAN(216.2000), max(396.0000).
        #
        #    (b2)skip-decel: count(5995), min(137.0000), MEAN(199.8333±78.0727), max(401.0000).
        #                    vehs: 02189=332, 02371=401m 0891=381, 00409=301
        #    Accel only errs count(1894), min(11.0000), MEAN(63.1333±63.211), max(220.0000)
        #
        #    Heinz-db Pwot:
        #       ±DIFFs: count(5758), min(135), MEAN(191.93±67.34), max(363).
        #       +DIFFs: count(1681), min(10), MEAN(56.03±54.33), max(192).
        #    Rule(f) before rule(bheinz_fname:
        #       ±DIFFs: count(5583), min(135), MEAN(186.10±57.99), max(335).
        #       +DIFFs: count(1515), min(10), MEAN(50.50±46.65), max(176).
        #    Rule(f) applied also forheinz_fnamei-2, i-3, ... signular-downshifts:
        #       ±DIFFs: count(5568), min(135), MEAN(185.60±57.59), max(335).
        #       +DIFFs: count(1500), min(10), MEAN(50.00±46.26), max(176).
        #    Precise input values:
        #       ±DIFFs: count(5569), min(135), MEAN(185.63±57.57), max(335).
        #       +DIFFs: count(1501), min(10), MEAN(50.03±46.25), max(176).
        #       ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #    Rule(e): Cancel +1 upshift for ANY previous-gear LOWER than the upshifted one:
        #       ±DIFFs: count(5547), min(135), MEAN(184.90±56.22), max(327).
        #       +DIFFs: count(1490), min(10), MEAN(49.67±45.68), max(173).
        #       ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #    Rule(e): Cancel ANY upshift for ANY previous-gear LOWER than the upshifted one:
        #       ±DIFFs: count(5533), min(135), MEAN(184.43±54.80), max(319).
        #       +DIFFs: count(1477), min(10), MEAN(49.23±44.40), max(161).
        #       ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #    Rule(g): Enable mistakenly impossible rule:
        #       ±DIFFs: count(5151), min(133), MEAN(171.70±41.96), max(288).
        #       +DIFFs: count(1116), min(8), MEAN(37.20±31.94), max(113).
        #       ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #    Rule(g): Apply only on Acell(not Flats):
        #       ±DIFFs: count(5149), min(133), MEAN(171.63±41.77), max(286).
        #       +DIFFs: count(1112), min(8), MEAN(37.07±31.73), max(112).
        #       ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #    Rule(b2): Apply it also for flats (not only Accels):
        #       ±DIFFs: count(5142), min(133), MEAN(171.40±41.15), max(279).
        #       +DIFFs: count(1109), min(8), MEAN(36.97±31.50), max(111).
        #       ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #    Rule(b2): Check if Accell on NEXT- time-step (A[t]).
        #           ±DIFFs: count(5275), min(133), MEAN(175.83±47.54), max(304).
        #           +DIFFs: count(1219), min(8), MEAN(40.63±36.33), max(128).
        #           ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #    Rule(b2): Check if Accell/FLAT on NEXT- time-step (A[t]).
        #           ±DIFFs: count(5306), min(135), MEAN(176.87±47.05), max(303).
        #           +DIFFs: count(1205), min(8), MEAN(40.17±35.84), max(127).
        #           ±ORIGs: count(539), min(7), MEAN(17.97±8.83), max(35).
        #
        # TRANSPLANTED:
        #       ±DIFFs: count(4771), min(122), MEAN(159.03±42.49), max(274).
        #       +DIFFs: count(752), min(3), MEAN(25.07±31.71), max(106).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(g) more greedy:
        #           ±DIFFs: count(4826), min(122), MEAN(160.87±41.65), max(268).
        #           +DIFFs: count(791), min(3), MEAN(26.37±30.30), max(101).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Apply Rule(g) on any beginning-gear (not on lower):
        #       ±DIFFs: count(4738), min(122), MEAN(157.93±37.54), max(255).
        #       +DIFFs: count(704), min(3), MEAN(23.47±26.32), max(88).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(g) on Accell but beggining on Accell OR flat:
        #       ±DIFFs: count(4676), min(122), MEAN(155.87±35.53), max(249).
        #       +DIFFs: count(642), min(3), MEAN(21.40±24.08), max(82).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    NO Rule(c)!
        #       ±DIFFs: count(1828), min(29), MEAN(60.93±37.66), max(157).
        #       +DIFFs: count(587), min(1), MEAN(19.57±24.41), max(81).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    NO RULE(d):             !!!(BUT err++ in ACCEL-phases)!!!
        #           ±DIFFs: count(1799), min(29), MEAN(59.97±36.93), max(156).
        #           +DIFFs: count(592), min(1), MEAN(19.73±24.54), max(81).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #     Overlapping application of rules:
        #       ±DIFFs: count(1826), min(29), MEAN(60.87±36.32), max(150).
        #       +DIFFs: count(576), min(1), MEAN(19.20±23.23), max(77).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    NO RULE(d):             !!!(BUT increase in ^^)!!!
        #           ±DIFFs: count(1800), min(29), MEAN(60.00±35.87), max(150).
        #           +DIFFs: count(582), min(1), MEAN(19.40±23.66), max(77).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    AFTER CHECK WITH HEINZ b reorder b1, b2 order:
        #       ±DIFFs: count(1739), min(29), MEAN(57.97±30.61), max(132).
        #       +DIFFs: count(539), min(1), MEAN(17.97±21.02), max(72).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    RENABLED Rule(c2): skip final 1st-gear.
        #       ±DIFFs: count(861), min(2), MEAN(28.70±30.43), max(105).
        #       +DIFFs: count(539), min(1), MEAN(17.97±21.02), max(72).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(c2) in the x2-loop:
        #       ±DIFFs: count(831), min(1), MEAN(27.70±30.43), max(104).
        #       +DIFFs: count(539), min(1), MEAN(17.97±21.02), max(72).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(a) after other rules:
        #       ±DIFFs: count(792), min(0), MEAN(26.40±30.53), max(103).
        #       +DIFFs: count(500), min(0), MEAN(16.67±21.15), max(71).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(g) slightly more greedy: accept any A in starting-T:
        #       ±DIFFs: count(481), min(0), MEAN(16.03±20.84), max(77).
        #       +DIFFs: count(234), min(0), MEAN(7.80±12.90), max(47).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(b2): check A>0 also for t-1 (not only t-2): But MAX++ & STDEV++
        #           ±DIFFs: count(479), min(0), MEAN(15.97±20.97), max(82).
        #           +DIFFs: count(231), min(0), MEAN(7.70±12.77), max(47).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(c1): check A<0 also for t-1 (not only t-2): But MAX++ & STDEV++
        #           ±DIFFs: count(474), min(0), MEAN(15.80±21.21), max(83).
        #           +DIFFs: count(230), min(0), MEAN(7.67±12.97), max(47).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    NO RULE(d):             !!!(BUT increase in ^^)!!!
        #           ±DIFFs: count(450), min(0), MEAN(15.00±20.59), max(76).
        #           +DIFFs: count(242), min(0), MEAN(8.07±13.59), max(47).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(c2): check A<0 also for t (not only t-1):
        #       ±DIFFs: count(476), min(0), MEAN(15.87±21.07), max(78).
        #       +DIFFs: count(233), min(0), MEAN(7.77±13.10), max(47).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    NO RULE(d):             !!!(BUT increase in ^^)!!!
        #           ±DIFFs: count(445), min(0), MEAN(14.83±20.82), max(77).
        #           +DIFFs: count(241), min(0), MEAN(8.03±13.82), max(48).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    First run rules(e & g) in isolation:
        #       ±DIFFs: count(546), min(0), MEAN(18.20±24.69), max(91).
        #       +DIFFs: count(258), min(0), MEAN(8.60±14.79), max(55).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    NO RULE(d):             !!!(BUT increase in ^^)!!!
        #       ±DIFFs: count(500), min(0), MEAN(16.67±23.40), max(87).
        #       +DIFFs: count(254), min(0), MEAN(8.47±14.61), max(55).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    First run rules(f) in isolation: !!!
        #       ±DIFFs: count(348), min(0), MEAN(11.60±15.83), max(61).
        #       +DIFFs: count(101), min(0), MEAN(3.37±5.82), max(25).heinz_fname        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(c1): Skip even further on decel.
        #       ±DIFFs: count(256), min(0), MEAN(8.53±13.40), max(48).
        #       +DIFFs: count(101), min(0), MEAN(3.37±5.82), max(25).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    First run rules(e) in isolation, g in main-loop:
        #       ±DIFFs: count(226), min(0), MEAN(7.53±11.74), max(52).
        #       +DIFFs: count(101), min(0), MEAN(3.37±5.41), max(24).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(c1) in isolation AFTER rest rules:([step_rule_g, step_rule_f][step_rule_e, step_rule_b1, step_rule_b2][step_rule_c1])
        #       ±DIFFs: count(214), min(0), MEAN(7.13±10.57), max(46).
        #       +DIFFs: count(97), min(0), MEAN(3.23±4.96), max(21).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(e) above rule b1&b2:
        #           ±DIFFs: count(211), min(0), MEAN(7.03±11.47), max(46).
        #           +DIFFs: count(93), min(0), MEAN(3.10±5.88), max(24).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    group rule(e & c1 (OR reversed)) ABOVE rules b1&b2:
        #           ±DIFFs: count(240), min(0), MEAN(8.00±12.87), max(53).
        #           +DIFFs: count(92), min(0), MEAN(3.07±5.83), max(24).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    group rule(e & c1 ) BELOW rules b1&b2:
        #           ±DIFFs: count(230), min(0), MEAN(7.67±11.78), max(46).
        #           +DIFFs: count(106), min(0), MEAN(3.53±5.57), max(21).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    group rule(e & c1 REVERSED) BELOW rules b1&b2:
        #           ±DIFFs: count(236), min(0), MEAN(7.87±12.46), max(52).
        #           +DIFFs: count(109), min(0), MEAN(3.63±5.90), max(24).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(f) also applied when G(t-1) GT_or_eq G(t+1) (the same also with apply-if-diff BUG fixed):
        #       ±DIFFs: count(187), min(0), MEAN(6.23±9.27), max(37).
        #       +DIFFs: count(72), min(0),heinz_fnameMEAN(2.40±3.76), max(13).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #        Ungroup rules(G, F):
        #           ±DIFFs: count(189), min(0), MEAN(6.30±9.42), max(37).
        #           +DIFFs: count(74), min(0), MEAN(2.47±3.88), max(13).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #        Ungroup rules(F, G):
        #           ±DIFFs: count(190), min(0), MEAN(6.33±9.48), max(37).
        #           +DIFFs: count(75), min(0), MEAN(2.50±3.96), max(13).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #        Rule(b2): check A>0 also till t (not t-1) (the same also with apply-if-diff BUG fixed):
        #           ±DIFFs: count(209), min(0), MEAN(6.97±11.04), max(46).
        #           +DIFFs: count(70), min(0), MEAN(2.33±3.65), max(12).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #        Rul(b2): Check A>0 for t-1 only (not only t-2 & t-1):
        #           ±DIFFs: count(202), min(0), MEAN(6.73±10.51), max(41).
        #           +DIFFs: count(64), min(0), MEAN(2.13±3.38), max(11).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #        Rule(f) applied on any singletton downshift:
        #           ±DIFFs: count(204), min(0), MEAN(6.80±9.23), max(32).
        #           +DIFFs: count(88), min(0), MEAN(2.93±3.78), max(12).
        #           ±ORIGs: count(0), miheinz_fname(0), MEAN(0.00±0.00), max(0).
        #        Rule(b2) Correct:
        #           ±DIFFs: count(209), min(0), MEAN(6.97±11.04), max(46).
        #           +DIFFs: count(70), min(0), MEAN(2.33±3.65), max(12).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #    Rule(b2) Correct NOT checking Accel on the final step of rule(b2)!:
        #       ±DIFFs: count(187), min(0), MEAN(6.23±9.27), max(37).
        #       +DIFFs: count(72), min(0), MEAN(2.40±3.76), max(13).
        #       ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #        Switch g with e, make e applied only for equal.
        #           ±DIFFs: count(284), min(0), MEAN(9.47±14.55), max(54).
        #           +DIFFs: count(127), min(0), MEAN(4.23±6.77), max(26).
        #           ±ORIGs: count(0), min(0), MEAN(0.00±0.00), max(0).
        #            Rule(f) applied on any singletton downshift:
        #               ±DIFFs: count(292), min(0), MEAN(9.73±14.50), max(49).
        #               +DIFFs: count(131), min(0), MEAN(4.37±6.60), max(25).
        #               ±ORIGs: count(0), min(0), heinz_fnameEAN(0.00±0.00), max(0).
        #    NO TRANSPLANT:
        #       ±DIFFs: count(772), min(7), MEAN(25.73±14.24), max(65).
        #       +DIFFs: count(499), min(2), MEAN(16.63±10.30), max(45).
        #       ±ORIGs: count(779), min(15), MEAN(25.97±8.83), max(43).
        #    Do not allow G2 when V==0:
        #       ±DIFFs: count(670), min(4), MEAN(22.33±13.88), max(61).
        #       +DIFFs: count(499), min(2), MEAN(16.63±10.30), max(45).
        #       ±ORIGs: count(719), min(13), MEAN(23.97±8.83), max(41).


    else:
        inpfname = os.path.join(mydir, samples_folder, 'sample_vehicles-{:05}.csv'.format(experiment_num))

        (df_my, df_hz, ndiff_gears, ndiff_gears_accel, ndiff_gears_orig)  = read_and_compare_experiment(inpfname, veh_num)

        ax = fig.axes()
        plotResults(os.path.basename(inpfname), df_my, df_hz, ndiff_gears, ax)
        log.info(">> %i: %s: ±DIFFs(%i), +DIFFs(%i), ±ORIGs(%i)", n, inpfname, ndiff_gears, ndiff_gears_accel, ndiff_gears_orig)


    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys

    heinz_dir               = None
    experiment_num          = None
    compare_original_gears  = False
    force_rerun_experiments = False  # Set to True to recalc experiments or 'compare_original_gears' has changed.
    try:
        if len(sys.argv) > 1:
            heinz_dir = sys.argv[1]

            if len(sys.argv) > 2:
                experiment_num = int(sys.argv[2])

                if len(sys.argv) > 3:
                    compare_original_gears = bool(sys.argv[3])
    except (ValueError, IndexError) as ex:
        exit('Help: \n  <cmd> [heinz_dir [vehicle_num]]\neg: \n  python experiment_SampleVehicleTests d:\Work/Fontaras\WLTNED\HeinzCycles\for_JRC_Petrol_* \nor \n  d:\Work/Fontaras\WLTNED\HeinzCycles\for_JRC_Petrol_*  2357')

    plot_diffs_with_heinz(heinz_dir, experiment_num=experiment_num, transplant_original_gears = compare_original_gears, force_rerun_experiments = force_rerun_experiments)



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
:created: 5 Jan 2014
'''

from ..experiment import Experiment
from .goodvehicle import goodVehicle
import logging
import unittest
import time


class ExperimentPerf(unittest.TestCase):

    @unittest.SkipTest
    def testPerf(self):
        logging.getLogger().setLevel(logging.WARNING)

        nexp = 100
        start = time.time()
        for _ in range(nexp):
            model = goodVehicle()

            experiment = Experiment(model)

            experiment.run()

        elapsed = (time.time() - start)
        print(">> ELAPSED: %.2fsec, RUN/EXP: %.4fsec"%(elapsed, elapsed/nexp))


if __name__ == "__main__":
    import sys;#sys.argv = ['', 'Test.testName']
    unittest.main(argv = sys.argv[1:])



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
:created: 5 Jan 2014
'''

from ..experiment import Experiment
from wltp import model
from .goodvehicle import goodVehicle
import unittest


class Test(unittest.TestCase):

    def testGoodVehicle(self):
        mdl = goodVehicle()

        exp = Experiment(mdl)
        mdl = exp.model
        self.assertEqual(mdl['vehicle']['full_load_curve'], model.default_load_curve())


    def testOverlayOnInit(self):
        mdl = goodVehicle()
        nval = 6000
        mdl2 = {
            "vehicle": {
                "n_rated":nval,
            }
        }

        exp = Experiment(mdl, mdl2)
        mdl = exp.model
        self.assertEqual(mdl['vehicle']['n_rated'], nval)

    def testMultiErrors(self):
        mdl = goodVehicle()
        mdl2 = {
            "vehicle": {
                "n_rated":-1,
                "n_idle":-2,
            }
        }

        exp = Experiment(mdl, mdl2, skip_model_validation=True)
        errors = list(exp.validate(True))
        self.assertIsNotNone(errors)
        self.assertEqual(len(errors), 2)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
:created: 5 Jan 2014
'''


from ..experiment import Experiment
from .goodvehicle import goodVehicle
from matplotlib import pyplot as plt
import logging
import numpy as np
import numpy.testing as npt
import os
import pickle
import tempfile
import unittest

log = logging.getLogger(__name__)

class ExperimentWholeVehs(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)
        self.run_comparison = True # NOTE: Set once to False to UPDATE sample-results (assuming they are ok).


    def compare_exp_results(self, tabular, fname, run_comparison):
        tmpfname = os.path.join(tempfile.gettempdir(), '%s.pkl'%fname)
        if (run_comparison):
            try:
                with open(tmpfname, 'rb') as tmpfile:
                    data_prev = pickle.load(tmpfile)
                    ## Compare changed-tabular
                    #
                    npt.assert_equal(tabular['gears'],  data_prev['gears'])
                    # Unreached code in case of assertion.
                    # cmp = tabular['gears'] != data_prev['gears']
                    # if (cmp.any()):
                    #     self.plotResults(data_prev)
                    #     print('>> COMPARING(%s): %s'%(fname, cmp.nonzero()))
                    # else:
                    #     print('>> COMPARING(%s): OK'%fname)
            except FileNotFoundError as ex:  # @UnusedVariable
                print('>> COMPARING(%s): No old-tabular found, 1st time to be stored in: '%fname, tmpfname)
                run_comparison = False

        if (not run_comparison):
            with open(tmpfname, 'wb') as tmpfile:
                pickle.dump(tabular, tmpfile)


    def plotResults(self, model):
        cycle = model['cycle_run']
        gears = cycle['gears']
        target = cycle['v_target']
        realv = cycle['v_real']
        clutch = cycle['clutch']

        clutch = clutch.nonzero()[0]
        plt.vlines(clutch,  0, 40)
        plt.plot(target)
        plt.plot(gears * 12, '+')
        plt.plot(realv)



    def testGoodVehicle(self, plot_results=False):
        logging.getLogger().setLevel(logging.DEBUG)

        model = goodVehicle()

        experiment = Experiment(model)
        model = experiment.run()
        self.assertTrue('cycle_run' in model, 'No result "cycle" in Model: %s'%model)

        print('DRIVEABILITY: \n%s' % experiment.driveability_report())
        cycle = model['cycle_run']
        gears = cycle['gears']
        print('G1: %s, G2: %s' % (np.count_nonzero(gears == 1), np.count_nonzero(gears == 2)))


        self.compare_exp_results(cycle, 'goodveh', self.run_comparison)


        if (plot_results):
            print(model['cycle_run'])
            #print([wltc_data()['classes']['class3b']['cycle_run'][k] for k in model['cycle_run']['driveability_issues'].keys()])
            self.plotResults(model)

            np.set_printoptions(edgeitems=16)
            #print(driveability_issues)
            #print(v_max)
            #results['target'] = []; print(results)
            plt.show()


    def testUnderPowered(self, plot_results=False):
        model = goodVehicle()
        model['vehicle']['p_rated'] = 50

        experiment = Experiment(model)
        model = experiment.run()
        print('DRIVEABILITY: \n%s' % experiment.driveability_report())
        self.compare_exp_results(model['cycle_run'], 'unpower1', self.run_comparison)


        model['vehicle']['mass']         =  1000
        model['vehicle']['p_rated']      =  80
        model['vehicle']['v_max']        =  120
        model['vehicle']['gear_ratios']  = [120.5, 95, 72, 52]

        experiment = Experiment(model)
        model = experiment.run()
        print('DRIVEABILITY: \n%s' % experiment.driveability_report())
        self.compare_exp_results(model['cycle_run'], 'unpower2', self.run_comparison)

        if (plot_results):
            self.plotResults(model)
            plt.show()


if __name__ == "__main__":
    import sys;#sys.argv = ['', 'Test.testName']
    unittest.main(argv = sys.argv[1:])



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
wltc module: WLTC gear-shift calculator
'''


from ._version import __version__
import semantic_version as semver


__author__          = "ankostis @ European Commission (JRC)"
__copyright__       = "Copyright (C) 2013-2014 European Commission (JRC)"
__license__         = "EUPL 1.1+"
__version_info__    = semver.Version(__version__)

__all__ = ['Experiment', 'model']



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
:created: 5 Jan 2014
'''


import logging
import unittest

import numpy as np

from .. import model
from wltp.experiment import (decideClass, calcPower_required, calcDownscaleFactor, downscaleCycle)


log = logging.getLogger(__name__)

class ExperimentDownscale(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)


    def testFull_manually(self):
        logging.getLogger().setLevel(logging.DEBUG)

        dtp                     = np.float64

        mass        = 1577.3106
        p_rated     = 78.6340
        v_max       = 186.4861
        f0          = 152.5813
        f1          = 307.5789
        f2          = 0.0486
        f_inertial  = 1.1 #constant
        ## Decide WLTC-class.
        #
        wltc                    = model.wltc_data()
        class_limits            = wltc['classification']['p_to_mass_class_limits']
        class3_velocity_split   = wltc['classification']['class3_split_velocity']
        wltc_class              = decideClass(class_limits, class3_velocity_split, mass, p_rated, v_max)
        class_data              = wltc['classes'][wltc_class]
        cycle                   = np.asarray(class_data['cycle'], dtype=dtp)

        A       = np.diff(cycle)
        A       = np.append(A, 0) # Restore element lost by diff().
        A       = A / 3.6
        P_REQ   = calcPower_required(cycle, A, mass, f0, f1, f2, f_inertial)

        f_downscale_threshold = 0.01
        dsc_data            = class_data['downscale']
        phases              = dsc_data['phases']
        p_max_values        = dsc_data['p_max_values']
        downsc_coeffs       = dsc_data['factor_coeffs']
        dsc_v_split         = dsc_data.get('v_max_split', None)
        f_downscale = calcDownscaleFactor(P_REQ, p_max_values, downsc_coeffs, dsc_v_split, p_rated, v_max, f_downscale_threshold)
        if (f_downscale > 0):
            cycle               = downscaleCycle(cycle, f_downscale, phases)


    def testCompareDownscaleMethodWithSpecs(self):
        from matplotlib import pyplot as plt
        '''Compare downcalings with the both methods: simplified (scale by multiply) and by_the_spec (iterativelly scale accelerations).'''

        dtp                     = np.float64

        wltc    = model.wltc_data()
        wclasses = wltc['classes']
        for wclass in wclasses.keys():
            for f_downscale in np.arange(0, 4, 0.1):
                class_data      = wclasses[wclass]
                cycle           = np.asarray(class_data['cycle'], dtype=dtp)
                phases          = class_data['downscale']['phases']
                V1              = downscaleCycle(cycle, f_downscale, phases)
                V2              = downscaleCycle_bySpecs(cycle, f_downscale, phases)

                err = np.abs(V1 - V2).max()
                print('Class(%s), f_dnscl(%s): %s'%(wclass, f_downscale, err))
                if (not np.allclose(V1, V2)):
                    print('Class(%s), f_dnscl(%s)'%(wclass, f_downscale))
                    plt.plot(cycle, 'r')
                    plt.plot(V1, 'b')
                    plt.plot(V2, 'g')
                    plt.show()
                    raise AssertionError('Class(%s), f_dnscl(%s)'%(wclass, f_downscale))


def downscaleCycle_bySpecs(cycle, f_downscale, phases):
    V = cycle.copy()
    (t0, t1, t2)    = phases

    ## Accelaration phase
    #
    for t in range(t0, t1):
        a       = cycle[t+1] - cycle[t]
        V[t+1]  = V[t] + a * (1-f_downscale)

    ## Decelaration phase
    #
    f_corr      = (V[t1] - cycle[t2]) / (cycle[t1] - cycle[t2])
    for t in range(t1, t2):
        a       = cycle[t+1] - cycle[t]
        V[t+1]  = V[t] + a * f_corr

    return V



if __name__ == "__main__":
    import sys;#sys.argv = ['', 'Test.testName']
    unittest.main(argv = sys.argv[1:])



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
:created: 5 Jan 2014
'''

import wltp.experiment as ex
from ..experiment import downscaleCycle
from ..model import wltc_data
import logging
import numpy as np
import numpy.testing as npt
import unittest


class experimentFuncs(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)

    def testDownscaling(self):
        wclasses = wltc_data()['classes']
        test_data = [(np.array(wclass['cycle']), wclass['downscale']['phases'], f_downscale)
                    for wclass in wclasses.values()
                    for f_downscale in np.linspace(0.1, 1, 10)]

        for (V, phases, f_downscale) in test_data:
            downscaleCycle(V, f_downscale, phases)


    def testNparray2Bytes(self):
        arr = np.array([-1, 0, 9, 10, 36, 255-ex._escape_char])

        self.assertEqual(ex.np2bytes(arr), b'\x7f\x80\x89\x8a\xa4\xff')
        self.assertRaisesRegex(AssertionError, 'Outside byte-range', ex.np2bytes, (arr + 1))
        self.assertRaisesRegex(AssertionError, 'Outside byte-range', ex.np2bytes, (arr - 1))

        npt.assert_array_equal(ex.bytes2np(ex.np2bytes(arr)), arr)


    def testRegex2bytes(self):
        regex = '\g1\g0\g24\g66\g127'

        self.assertEqual(ex.gearsregex(regex).pattern,  b'\x81\x80\x98\xc2\xff')

        regex = '\g1\g0|\g24\g66\g127'

        self.assertEqual(ex.gearsregex(regex).pattern,  b'\x81\x80|\x98\xc2\xff')


if __name__ == "__main__":
    import sys
    unittest.main(verbosity=2, argv=sys.argv)



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''The hierarchical data and their schema for the WLTC calculator (defaults, WLTC-data) used by the Model and Experiments classes.
'''

import numpy as np
from wltp.cycles import (class1, class2, class3)


def model_base():
    '''The base model for running a WLTC experiment.

    It contains some default values for the experiment (ie the default 'full-load-curve' for the vehicles).
    But note that it this model is not valid - you need to iverride its attributes.

    :return :json_tree: with the default values for the experiment.
    '''

    n_norm = np.arange(0.0, 1.21, 0.01)
    ## PETROL:
    default_load_curve_petrol = [
#        np.polyval([-1.0411, 1.3853, -0.5647, 1.1107, 0.0967], n_norm).tolist()
        n_norm.tolist(),
        ## Form Heinz-db
        [
        0.1       ,  0.11100069,  0.12200138,  0.13300206,  0.14400275,
        0.15500344,  0.16600413,  0.17700482,  0.1880055 ,  0.19900619,
        0.21000688,  0.22100757,  0.23200826,  0.24300895,  0.25400963,
        0.26501032,  0.27506528,  0.28512023,  0.29517519,  0.30523015,
        0.3152851 ,  0.3231181 ,  0.3309511 ,  0.3387841 ,  0.3466171 ,
        0.3544501 ,  0.3622831 ,  0.37011609,  0.37794909,  0.38578209,
        0.39361509,  0.40268373,  0.41175238,  0.42082102,  0.42988967,
        0.43895831,  0.44802696,  0.4570956 ,  0.46616424,  0.47523289,
        0.48430153,  0.49385337,  0.5034052 ,  0.51295704,  0.52250887,
        0.53206071,  0.54161255,  0.55116438,  0.56071622,  0.57026805,
        0.57981989,  0.58873474,  0.59764959,  0.60656444,  0.61547929,
        0.62439414,  0.633309  ,  0.64222385,  0.6511387 ,  0.66005355,
        0.6689684 ,  0.67815415,  0.6873399 ,  0.69652565,  0.70571139,
        0.71489714,  0.72408289,  0.73326864,  0.74245439,  0.75164013,
        0.76082588,  0.77026954,  0.77971321,  0.78915687,  0.79860054,
        0.8080442 ,  0.81748786,  0.82693153,  0.83637519,  0.84581885,
        0.85526252,  0.86393898,  0.87261544,  0.8812919 ,  0.88996837,
        0.89864483,  0.90732129,  0.91599775,  0.92467422,  0.93335068,
        0.94202714,  0.94782443,  0.95362171,  0.959419  ,  0.96521628,
        0.97101357,  0.97681086,  0.98260814,  0.98840543,  0.99420271,
        1.        ,  0.99556631,  0.99113261,  0.98669892,  0.98226523,
        0.97783154,  0.97339784,  0.96896415,  0.96453046,  0.96009677,
        0.95566307,  0.94509677,  0.93453046,  0.92396415,  0.91339784,
        0.90283154,  0.89226523,  0.88169892,  0.87113261,  0.86056631,
        0.85
        ]
    ]

    ## DIESEL:
    default_load_curve_diesel = [
#        np.polyval([-0.909, 1.9298, -2.2212, 2.088, 0.095], n_norm).tolist()
        n_norm.tolist(),
        ## Form Heinz-db
        [
         0.1       ,  0.11632768,  0.13265536,  0.14898304,  0.16531072,
        0.1816384 ,  0.19796609,  0.21429377,  0.23062145,  0.24694913,
        0.26327681,  0.27956221,  0.29584762,  0.31213302,  0.32841843,
        0.34470383,  0.36098924,  0.37727464,  0.39356004,  0.40984545,
        0.42613085,  0.44289144,  0.45965203,  0.47641262,  0.49317321,
        0.5099338 ,  0.52669439,  0.54345498,  0.56021557,  0.57697616,
        0.59373675,  0.60775731,  0.62177786,  0.63579841,  0.64981897,
        0.66383952,  0.67786007,  0.69188063,  0.70590118,  0.71992173,
        0.73394229,  0.74334518,  0.75274808,  0.76215098,  0.77155387,
        0.78095677,  0.79035967,  0.79976256,  0.80916546,  0.81856836,
        0.82797126,  0.83355082,  0.83913039,  0.84470996,  0.85028953,
        0.8558691 ,  0.86144867,  0.86702823,  0.8726078 ,  0.87818737,
        0.88376694,  0.88861739,  0.89346785,  0.8983183 ,  0.90316876,
        0.90801921,  0.91286967,  0.91772013,  0.92257058,  0.92742104,
        0.93227149,  0.93551314,  0.93875479,  0.94199644,  0.94523809,
        0.94847974,  0.95172139,  0.95496304,  0.95820469,  0.96144634,
        0.96468798,  0.96645067,  0.96821335,  0.96997604,  0.97173872,
        0.97350141,  0.97526409,  0.97702678,  0.97878946,  0.98055215,
        0.98231483,  0.98408335,  0.98585187,  0.98762038,  0.9893889 ,
        0.99115742,  0.99292593,  0.99469445,  0.99646297,  0.99823148,
        1.        ,  0.98871106,  0.97742212,  0.96613319,  0.95484425,
        0.94355531,  0.93226637,  0.92097743,  0.9096885 ,  0.89839956,
        0.88711062,  0.88339956,  0.8796885 ,  0.87597743,  0.87226637,
        0.86855531,  0.86484425,  0.86113319,  0.85742212,  0.85371106,
        0.85
        ]
    ]

    instance = {
        'vehicle': {
            "mass":     None,
            "v_max":    None,
            "p_rated":  None,
            "n_rated":  None,
            "n_idle":   None,
            "n_min":    None,
            "gear_ratios":[],
            "resistance_coeffs":[],
            'full_load_curve': default_load_curve_petrol, # FIXME: Decide load_curtve by engine-type!
        },
        'params': {
            'v_stopped_threshold':      1,          # Km/h, <=
            'f_inertial':               1.1,
            'f_safety_margin':          0.9,
            'f_n_max':                  1.2,
            'f_n_min':                  0.125,
            'f_n_min_gear2':            0.9,
            'f_n_clutch_gear2':         [1.15, 0.03],
        }
    }

    return instance


def default_vehicle():
    return model_base()['vehicle']

def default_load_curve():
    return default_vehicle()['full_load_curve']


def results_base():
    instance = {
        'wltc_class': None,
        'v_target': [],
        'gears': [],
        'v_real': [],
        'downscale_factor': None,
        'driveability': None,
    }

    return instance


def wltc_data():
    '''The WLTC-data required to run an experiment (the class-cycles and their attributes)..

    :return :json_tree:
    '''

    ## See schemas for explainations.
    ##
    wltc_data = {
        'classes': {
            'class1': class1.class_data(),
            'class2': class2.class_data(),
            'class3a': class3.class_data_a(),
            'class3b': class3.class_data_b(),
        },
        'classification': {
            'p_to_mass_class_limits':   [22, 34], #  W/kg, <=, choose class1/2/3
            'class3_split_velocity':    120,       # Km/h, <, choose class3a/3b
        }
    }

    return wltc_data


def merge(a, b, path=[]):
    ''''merges b into a'''

    from collections.abc import Mapping

    for key in b:
        bv = b[key]
        if key in a:
            av = a[key]
            if isinstance(av, Mapping) != isinstance(bv, Mapping):
                raise ValueError("Dict-values conflict at '%s'! a(%s) != b(%s)" %
                                ('/'.join(path + [str(key)]), type(av), type(bv)))
            elif av == bv:
                continue # same leaf value
            elif isinstance(av, Mapping):
                merge(av, bv, path + [str(key)])
                continue
        a[key] = bv
    return a

# works
# print(merge({1:{"a":"A"},2:{"b":"B"}}, {2:{"c":"C"},3:{"d":"D"}}))
# # has conflict
# merge({1:{"a":"A"},2:{"b":"B"}}, {1:{"a":"A"},2:{"b":"C"}})




def model_schema():
    '''The json-schema for input/output of the WLTC experiment.

    :return :dict:
    '''

    from textwrap import dedent

    schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'title': 'Json-schema describing the input for a WLTC experiment.',
        'type': 'object', 'additionalProperties': False,
        'required': ['vehicle'],
        'properties': {
            'vehicle': {
                'title': 'vehicle model',
                'type': 'object', 'additionalProperties': False,
                'required': ['mass', 'v_max', 'p_rated', 'n_rated', 'n_idle', 'gear_ratios', 'resistance_coeffs', 'full_load_curve'],
                'description': 'The vehicle attributes required for generating the WLTC velocity-profile downscaling and gear-shifts.',
                'properties': {
                   'mass': {
                       'title': 'vehicle test mass',
                       '$ref': '#/definitions/positiveNumber',
                       'description': 'The test mass of the vehicle in kg.',
                    },
                   'v_max': {
                       'title': 'maximum vehicle velocity',
                       'type': ['integer', 'null'],
                       'minimum': 0,
                       'exclusiveMinimum': True,
                       'description': dedent('''
                           The maximum velocity as declared by the manufacturer.
                           If ommited, calculated as::
                               v_max = (n_rated * f_n_max (=1.2)) / gear_ratio[last]
                       '''),
                   },
                   'p_rated': {
                       'title': 'maximum rated power',
                       '$ref': '#/definitions/positiveNumber',
                       'description': 'The maximum rated engine power as declared by the manufacturer.',
                   },
                   'n_rated': {
                       'title': 'rated engine revolutions',
                       '$ref': '#/definitions/positiveNumber',
                       'description': dedent('''
                           The rated engine revolutions at which an engine develops its maximum power.
                           If the maximum power is developed over an engine revolutions range,
                           it is determined by the mean of this range.
                           This is called 's' in the specs.
                       '''),
                    },
                   'n_idle': {
                       'title': 'idling revolutions',
                       '$ref': '#/definitions/positiveNumber',
                       'description': 'The idling engine revolutions as defined of Annex 1.',
                    },
                   'n_min': {
                       'title': 'minimum engine revolutions',
                       'type': ['integer', 'null'],
                       'description': dedent('''
                        minimum engine revolutions for gears > 2 when the vehicle is in motion. The minimum value
                        is determined by the following equation::
                            n_min = n_idle + f_n_min(=0.125) * (n_rated - n_idle)
                        Higher values may be used if requested by the manufacturer, by setting this one.
                       '''),
                    },
                   'gear_ratios': {
                       'title': 'gear ratios',
                       '$ref': '#/definitions/positiveNumbers',
                       'maxItems': 24,
                       'minItems': 3,
                       'description':
                       'An array with the gear-ratios obtained by dividing engine-revolutions (1/min) by vehicle-velocity (km/h).',
                    },
                   'resistance_coeffs': {
                       'title': 'driving resistance coefficients',
                       'type': 'array', 'items': {'type': 'number'},
                       'minItems': 3,
                       'maxItems': 3,
                       'description': 'The 3 driving resistance coefficients f0, f1, f2 as defined in Annex 4, in N, N/(km/h), and N/(km/h)² respectively.',
                    },
                   'full_load_curve': {
                       'title': 'full load power curve',
                       'type': 'array',
                       'items': [
                            {
                               'title': 'normalized engine revolutions',
                               'description': dedent('''
                                    The normalized engine revolutions, within [0.0, 0.15]::
                                        n_norm = (n - n_idle) / (n_rated  - n_idle)
                                    '''),
                               'type': 'array', 'additionalItems': False,
                               'maxItems': 360,
                               'minItems': 7,
                               'items': {
                                   'type': 'number',
                                   'minimum': 0.0,
                                   'exclusiveMinimum': False,
                                   'maximum': 1.5,
                                   'exclusiveMaximum': False,
                                },
                            },
                            {
                               'title': 'normalized full-load power curve',
                               'description': dedent('''
                                    The normalised values of the full-power load against the p_rated,
                                    within [0, 1]::
                                        p_norm = p / p_rated
                                '''),
                               'type': 'array', 'additionalItems': False,
                               'maxItems': 360,
                               'minItems': 7,
                               'items': {
                                   'type': 'number',
                                   'minimum': 0.0,
                                   'exclusiveMinimum': False,
                                   'maximum': 1.0,
                                   'exclusiveMaximum': False,
                                }
                            },
                        ],
                       'description': dedent('''
                            A 2-dimensional array holding the full load power curve in 2 rows
                            Example::
                                [
                                    [ 0, 10, 20, 30, 40, 50, 60, 70. 80, 90 100, 110, 120 ],
                                    [ 6.11, 21.97, 37.43, 51.05, 62.61, 72.49, 81.13, 88.7, 94.92, 98.99, 100., 96.28, 87.66 ]
                                ]

                       '''),
                    },
                }  #veh-props
            }, # veh
            'params': {
                'title': 'experiment parameters',
                'type': 'object', 'additionalProperties': False,
                'required': [
                    'v_stopped_threshold',
                    'f_inertial',
                    'f_safety_margin',
                    'f_n_max',
                    'f_n_min',
                    'f_n_min_gear2',
                    'f_n_clutch_gear2',
                ],
                'properties': {
                    'v_stopped_threshold': {
                        'description': 'Velocity (Km/h) under which (<=) to idle gear-shift (Annex 2-3.3, p71).',
                        'type': [ 'number', 'null'],
                        'default': 1,
                    },
                    'f_inertial': {
                        'description': 'Inertial factor used for calculating required-power (Annex 2-3.1, p71).',
                        'type': [ 'number', 'null'],
                        'default': 1.1,
                    },
                    'f_safety_margin': {
                        'description': 'Safety-margin factor for load-curve due to transitional effects (Annex 2-3.3, p72).',
                        'type': [ 'number', 'null'],
                        'default': 0.9,
                    },
                    'f_n_max': {
                        'description': 'For each gear, N :< n_max = n_idle + f_n_max * n_range',
                        'type': [ 'number', 'null'],
                        'default': 1.2,
                    },
                    'f_n_min': {
                        'description': 'For each gear > 2, N :> n_min = n_idle + f_n_min * n_range (unless ``n_min`` overriden by manufacturer)',
                        'type': [ 'number', 'null'],
                        'default': 0.125,
                    },
                    'f_n_min_gear2': {
                        'description': 'Gear-2 is invalid when N :< f_n_min_gear2 * n_idle.',
                        'type': [ 'number', 'null'],
                        'default': 0.9,
                    },
                    'f_n_clutch_gear2': {
                        'description': dedent('''
                            A 2-value number-array(f1, f2) controlling when to clutch gear-2::
                                N < n_clutch_gear2 := max(f1 * n_idle, f2 * n_range + n_idle),
                            unless "clutched"...
                        '''),
                        'type': [ 'array', 'null'],
                        'default': [1.15, 0.03],
                    },
                    'wltc_class': {
                        'description': 'The name of the WLTC-class (found within WLTC-data/classes) as selected by the experiment.',
                        'type': 'string',
                    },
                    'f_downscale': {
                        'description': 'The downscaling-factor as calculated by the experiment (Annex 1-7.3, p68).',
                        'type': 'number',
                    },
                }
            },
            'cycle_run': {}, #TODO: results(cycle) model-schema.
        },
        'definitions': {
            'positiveInteger': {
                'type': 'integer',
                'minimum': 0,
                'exclusiveMinimum': True,
            },
            'positiveNumber': {
                'type': 'number',
                'minimum': 0,
                'exclusiveMinimum': True,
            },
            'positiveIntegers': {
                'type': 'array',
               'items': { '$ref': '#/definitions/positiveInteger' },
            },
            'positiveNumbers': {
                'type': 'array',
               'items': { '$ref': '#/definitions/positiveNumber' },
            },
            'mergeableArray': {
                'type': 'object', '': False,
                'required': ['$merge', '$list'],
                'properties': {
                    '$merge': {
                        'enum': ['merge', 'replace', 'append_head', 'append_tail', 'overwrite_head', 'overwrite_tail'],
                        'description': dedent('''
                            merge       := appends any non-existent elements
                            replace     := (default) all items replaced
                        '''),
                    },
                    '$list': {
                        'type': 'array',
                    }
                },
            },
            'mergeableObject': {
                'type': 'object',
                'properties': {
                    '$merge': {
                        'type': 'boolean',
                        'description': dedent('''
                            true    := (default) merge properties
                            false   := replace properties
                        '''),
                    },
                },
            },
        }
    }

    return schema

def wltc_schema():
    '''The json-schema for the WLTC-data required to run a WLTC experiment.

    :return :dict:
    '''

    schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'title': 'WLTC data',
        'type': 'object', 'additionalProperties': False,
        'required': ['classes', 'classification'],
        'properties': {
            'classes': {
                'type': 'object', 'additionalProperties': False,
                'required': ['class1', 'class2', 'class3a', 'class3b'],
                'properties': {
                    'class1': {'$ref': '#/definitions/class'},
                    'class2': {'$ref': '#/definitions/class'},
                    'class3a': {'$ref': '#/definitions/class'},
                    'class3b': {'$ref': '#/definitions/class'},
                }
            },
            'classification': {
                'type': 'object', 'additionalProperties': False,
                'required': ['p_to_mass_class_limits', 'class3_split_velocity'],
                'properties': {
                    'p_to_mass_class_limits': {
                        'description': 'Power_to_unladen-mass ratio (W/kg) used to select class (Annex 1, p19).',
                        'type': 'array',
                        'default': [22, 34],
                    },
                    'class3_split_velocity': {
                        'description': 'Velocity (Km/h) under which (<) version-B from class3 is selected (Annex 1, p19).',
                        'type': 'integer',
                        'default':120,
                    },
                }
            },
        },
        'definitions': {
            'class': {
                'title': 'WLTC class data',
                'type': 'object', 'additionalProperties': False,
                'required': ['parts', 'downscale', 'cycle'],
                'properties': {
                    'parts': {
                        'type': 'array', 'items': {
                            'type': 'array', 'items': {'type': 'integer',},
                            'minItems': 2,
                            'maxItems': 2,
                        }
                    },
                    'downscale': {
                        'type': 'object', 'additionalProperties': False,
                        'required': ['phases', 'p_max_values', 'factor_coeffs'],
                        'properties': {
                            'phases': {
                                'type': 'array', 'additionalItems': False,
                                'items': {
                                    'type': 'integer'
                                },
                                'maxItems': 3, 'minItems': 3,
                            },
                            'decel_phase': {
                                'type': 'array', 'additionalItems': False,
                                'items': {'type': 'integer'},
                                'maxItems': 2, 'minItems': 2,
                            },
                            'p_max_values': {
                                'type': 'array', 'additionalItems': False,
                                'items': { 'type': 'number'},
                                'maxItems': 3, 'minItems': 3,
                            },
                            'factor_coeffs': {
                                'type': 'array',
                            },
                            'v_max_split': {
                                'type': 'number',
                            },
                        }
                    },
                    'cycle': {
                        'type': 'array',
                        'items': {'type': 'number',},
                        'minItems': 906, # class1's length
                    }
                },
            }
        }
    }

    return schema


def model_validator():
    from jsonschema import Draft4Validator
    schema = model_schema()
    return Draft4Validator(schema)

def wltc_validator():
    from jsonschema import Draft4Validator
    schema = wltc_schema()
    return Draft4Validator(schema)


def validate_full_load_curve(flc, f_n_max):
    if (min(flc[0]) > 0):
        raise ValueError('The full_load_curve must begin at least from 0%%, not from %f%%!' % min(flc[0]))
    max_x_limit = f_n_max
    if (max(flc[0]) < max_x_limit):
        raise ValueError('The full_load_curve must finish at least on f_n_max(%f%%), not on %f%%!' % (max_x_limit, max(flc[0])))


if __name__ == '__main__':
    import json
    print('Model: %s' % json.dumps([model_schema(), wltc_schema()], indent=1))
    print('Model: %s' % json.dumps(model_base(), indent=1))




#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
''''Setuptools script for *wltp*, the WLTC gear-shift calculator.

Install:
========

To install it, assuming you have download the sources,
do the usual::

    python setup.py install

Or get it directly from the PIP repository::

    pip3 install wltp
'''
## Got ideas for project-setup from many places, among others:
#    http://www.jeffknupp.com/blog/2013/08/16/open-sourcing-a-python-project-the-right-way/
#    http://python-packaging-user-guide.readthedocs.org/en/latest/current.html

from setuptools import setup
#from cx_Freeze import Executable
#from cx_Freeze import setup
import os

projname = 'wltp'
mydir = os.path.dirname(__file__)

## Version-trick to have version-info in a single place,
## taken from: http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
##
def read_project_version():
    fglobals = {}
    exec(open(os.path.join(mydir, projname, '_version.py')).read(), fglobals)  # To read __version_info__
    return fglobals['__version__']
proj_ver = read_project_version()

def read_text_lines(fname):
    with open(os.path.join(mydir, fname)) as fd:
        return fd.readlines()

readme_lines = read_text_lines('README.rst')

long_desc = readme_lines + ['\n\n\n'] + read_text_lines('CHANGES.rst')
long_desc = '\n'.join(long_desc)

setup(
    name = projname,
    packages = ['wltp', 'wltp.cycles', 'wltp.test', ],
#     package_data= {'projname': ['data/*.csv']},
#     scripts = ['wltp.py'],
    version=proj_ver,
    test_suite="wltp.test", #TODO: check setup.py testsuite indeed works.
    description=readme_lines[1],
    long_description=long_desc,
    author="Kostis Anagnostopoulos @ European Commission (JRC)",
    author_email="ankostis@gmail.com",
    url = "https://github.com/ankostis/wltp",
    license = "European Union Public Licence 1.1 or later (EUPL 1.1+)",
    keywords = [
        "LDVs", "UN", "UNECE", "cars", "fuel-consumption", "simulation", "simulator",
        "gears", "gearshifs", "rpm", "vehicles", "wltc", "wltp"
    ],
    classifiers = [
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        'Natural Language :: English',
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Manufacturing",
        "License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires = [
        'semantic_version',
        'numpy',
        'jsonschema',
#         'jsonpointer',
#         'pandas',
#         'pint',
    ],
    options = {
        'build_exe': {
            "excludes": [
                "jsonschema", #!!!! Schemas do not work in library-zip, so needs manuall to copy directly into app-dir
                "numpy", "scipy", #!!!! lostesso
                "PyQt4", "PySide",
                "IPython", "numexpr",
                "pygments", "pyreadline", "jinja2",
                "setuptools",
                "statsmodels", "docutils",
                "xmlrpc", "pytz",
                "nose",
                "Cython", "pydoc_data", "sphinx", "docutils",
                "multiprocessing", "lib2to3", "_markerlib",
#                 #urllib<--email<--http<--pandas
#                 #distutils" <-- pandas.compat
            ],
            'includes': [
#                 'matplotlib.backends.backend_tkagg',
            ],
            'include_files': [
                ## MANUAL COPY into build/exe-dir
                ##     from: https://bitbucket.org/anthony_tuininga/cx_freeze/issue/43/import-errors-when-using-cx_freeze-with
                #  site_packages(32bit/64bit)/
                #    jsonschema
                #    numpy
                #    scipy
            ],
        }, 'bdist_msi': {
            'add_to_path': False,
        },
    },
#     executables=[Executable("wltp.py", )], #base="Win32GUI")],
)



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
:created: 5 Jan 2014
'''

def goodVehicle():
    goodVehicle = {
        "vehicle": {
            "mass":     1500,
            "v_max":    None,
            "p_rated":  100,
            "n_rated":  5450,
            "n_idle":   950,
            #"n_min":   None,    # Can be overriden by manufacturer.
            "gear_ratios":      [120.5, 75, 50, 43, 37, 32],
            "resistance_coeffs":[100, 0.5, 0.04],
        }
    }
    return goodVehicle



#!/usr/bin/env python
#
# Copyright 2013-2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''Testing of the pure-tree data (just dictionary & lists), without the Model/Experiment classes.

:created: 29 Dec 2013
'''

from .. import model
import json
import jsonschema
import unittest


class Test(unittest.TestCase):

    def setUp(self):
        self.goodVehicle_jsonTxt = '''{"vehicle": {
            "mass":1300,
            "v_max":195,
            "p_rated":110.625,
            "n_rated":5450,
            "n_idle":950,
            "n_min":500,
            "gear_ratios":[120.5, 75, 50, 43, 33, 28],
            "resistance_coeffs":[100, 0.5, 0.04]
            %s
        }}'''


    def testWltcData(self):
        mdl = model.wltc_data()
        validator = model.wltc_validator()

        validator.validate(mdl)


    def testModelBase(self):
        mdl = model.model_base()

        self.assertRaises(jsonschema.ValidationError, model.model_validator().validate, mdl)


    def testModelInstance_missingLoadCurve(self):
        json_txt = self.goodVehicle_jsonTxt % ('')
        mdl = json.loads(json_txt)
        validator = model.model_validator()

        self.assertRaisesRegex(jsonschema.ValidationError, "'full_load_curve' is a required property", validator.validate, mdl)


    def testModelInstance_simplInstanceeFullLoadCurve(self):
        json_txt = self.goodVehicle_jsonTxt % \
            (', "full_load_curve":[[ 1,  1,  1,  1,  1,  1,  1,  1,  1],   [ 0.23,  0.23, 0.23, 0.23, 0.23, 0.23, 0.23, 0.23, 0.23]]')
        mdl = json.loads(json_txt)

        model.model_validator().validate(mdl)
        self.assertNotEqual(mdl['vehicle']['full_load_curve'], model.default_load_curve())


    def testModelInstance_defaultLoadCurve(self):
        json_txt = self.goodVehicle_jsonTxt % \
            (', "full_load_curve":%s' % (json.dumps(model.default_load_curve())))
        mdl = json.loads(json_txt)
        validator = model.model_validator()

        validator.validate(mdl)



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
